r"""
Sonora Spectrum
-----------------

A container for a single grid-point of the Sonora precomputed synthetic model spectrum of brown dwarfs and free-floating Gas Giant planets.  The spectrum is a vector with coordinates wavelength and flux :math:`F(\lambda)`.

SonoraSpectrum
###############
"""

import copy
import warnings
import logging
from gollum.precomputed_spectrum import PrecomputedSpectrum
from gollum.telluric import TelluricSpectrum
import numpy as np
import astropy
import pandas as pd
from astropy import units as u
from specutils import SpectrumCollection
import os
from specutils.spectra.spectrum1d import Spectrum1D
from tqdm import tqdm

from math import sqrt

from bokeh.io import show, output_notebook, push_notebook
from bokeh.plotting import figure, ColumnDataSource
from bokeh.models import Slider, Span, Range1d, Dropdown
from bokeh.layouts import layout, Spacer
from bokeh.models.widgets import Button, Div

log = logging.getLogger(__name__)

#  See Issue: https://github.com/astropy/specutils/issues/779
warnings.filterwarnings(
    "ignore", category=astropy.utils.exceptions.AstropyDeprecationWarning
)
# See Issue: https://github.com/astropy/specutils/issues/800
warnings.filterwarnings("ignore", category=RuntimeWarning)


class Sonora2017Spectrum(PrecomputedSpectrum):
    """
    A container for a single Sonora precomputed synthetic spectrum of a brown dwarfs or free-floating
    Gas Giant planet.

    Args:
        Teff (int): The Teff label of the Sonora model to read in.  Must be on the Sonora grid.
        logg (float): The logg label of the Sonora model to read in.  Must be on the Sonora grid.
        path (str): The path to your local Sonora grid library.  You must have the Sonora grid downloaded locally.  Default: "~/libraries/raw/Sonora/"
        wl_lo (float): the bluest wavelength of the models to keep (Angstroms)
        wl_hi (float): the reddest wavelength of the models to keep (Angstroms)
        """

    def __init__(
        self, *args, teff=None, logg=None, path=None, wl_lo=8038, wl_hi=12849, **kwargs
    ):

        teff_points = np.hstack(
            (
                np.arange(500, 600, 25),
                np.arange(600, 1000, 50),
                np.arange(1000, 2401, 100),
            )
        )
        logg_points = np.arange(4.0, 5.51, 0.25)

        # Map logg (cgs) to the gravity labels used in file names
        logg_par_dict = {
            4.0: "100",
            4.25: "178",
            4.5: "316",
            4.75: "562",
            5.0: "1000",
            5.25: "1780",
            5.5: "3160",
        }

        if path is None:
            path = "~/libraries/raw/Sonora/"

        if (teff is not None) & (logg is not None):
            base_path = os.path.expanduser(path)
            assert os.path.exists(
                base_path
            ), "You must specify the path to local Sonora models: {}".format(base_path)

            print("successfully found the path to Sonora models: {}".format(base_path))

            assert teff in teff_points, "Teff must be on the grid points"
            assert logg in logg_points, "logg must be on the grid points"

            base_name = "sp_t{0:0>.0f}g{1:}nc_m0.0".format(
                float(teff), logg_par_dict[logg]
            )
            fn = base_path + "/" + base_name + ".gz"

            assert os.path.exists(fn), "Double check that the file {} exists".format(fn)

            # Units: micron, erg/cm^2/s/Hz
            df_native = (
                pd.read_csv(
                    fn,
                    skiprows=[0, 1],
                    delim_whitespace=True,
                    compression="gzip",
                    names=["wavelength_um", "flux"],
                )
                .sort_values("wavelength_um")
                .reset_index(drop=True)
            )

            # convert to Angstrom
            df_native["wavelength"] = df_native["wavelength_um"] * 10_000.0
            mask = (df_native.wavelength > wl_lo) & (df_native.wavelength < wl_hi)
            df_trimmed = df_native[mask].reset_index(drop=True)

            super().__init__(
                spectral_axis=df_trimmed.wavelength.values * u.Angstrom,
                flux=df_trimmed.flux.values * u.erg / u.s / u.cm ** 2 / u.Hz,
                **kwargs,
            )

        else:
            super().__init__(*args, **kwargs)


class Sonora2021Spectrum(PrecomputedSpectrum):
    r"""
    A container for a single Sonora precomputed synthetic spectrum of a brown dwarfs or free-floating
    Gas Giant planet.

    Args:
        Teff (int): The Teff label of the Sonora model to read in.  Must be on the Sonora grid.
        logg (float): The logg label of the Sonora model to read in.  Must be on the Sonora grid.
        path (str): The path to your local Sonora grid library.  You must have the Sonora grid downloaded locally.  Default: "~/libraries/raw/Sonora/"
        wl_lo (float): the bluest wavelength of the models to keep (Angstroms)
        wl_hi (float): the reddest wavelength of the models to keep (Angstroms)
        """

    def __init__(
        self,
        *args,
        teff=None,
        logg=None,
        metallicity=0.0,
        path=None,
        wl_lo=8038,
        wl_hi=12849,
        **kwargs,
    ):

        teff_points = np.hstack(
            (
                np.arange(500, 600, 25),
                np.arange(600, 1000, 50),
                np.arange(1000, 2401, 100),
            )
        )
        logg_points = np.arange(3.0, 5.51, 0.25)

        # Map logg (cgs) to the gravity labels used in file names
        logg_par_dict = {
            3.0: "10",
            3.25: "17",
            3.5: "31",
            3.75: "56",
            4.0: "100",
            4.25: "178",
            4.5: "316",
            4.75: "562",
            5.0: "1000",
            5.25: "1780",
            5.5: "3160",
        }

        metallicity_points = np.arange(-0.5, 0.51, 0.5)

        if path is None:
            path = "~/libraries/raw/SonoraBobcat2021/"

        if (teff is not None) & (logg is not None):
            base_path = os.path.expanduser(path)
            assert os.path.exists(
                base_path
            ), "You must specify the path to local Sonora models: {}".format(base_path)

            assert teff in teff_points, "Teff must be on the grid points"
            assert logg in logg_points, "logg must be on the grid points"
            assert metallicity in metallicity_points, "Fe/H must be a valid point"

            if metallicity <= 0:
                base_name = "sp_t{0:0>.0f}g{1:}nc_m{2:0.01f}".format(
                    float(teff), logg_par_dict[logg], float(metallicity)
                )
            else:
                base_name = "sp_t{0:0>.0f}g{1:}nc_m{2:+0.1f}".format(
                    float(teff), logg_par_dict[logg], float(metallicity)
                )
            fn = base_path + "/" + base_name

            assert os.path.exists(fn), "Double check that the file {} exists".format(fn)

            # Units: micron, erg/cm^2/s/Hz
            df_native = (
                pd.read_csv(
                    fn,
                    skiprows=[0, 1],
                    delim_whitespace=True,
                    names=["wavelength_um", "flux"],
                )
                .sort_values("wavelength_um")
                .reset_index(drop=True)
            )

            # convert to Angstrom
            df_native["wavelength"] = df_native["wavelength_um"] * 10_000.0
            mask = (df_native.wavelength > wl_lo) & (df_native.wavelength < wl_hi)
            df_trimmed = df_native[mask].reset_index(drop=True)

            super().__init__(
                spectral_axis=df_trimmed.wavelength.values * u.Angstrom,
                flux=df_trimmed.flux.values * u.erg / u.s / u.cm ** 2 / u.Hz,
                **kwargs,
            )

        else:
            super().__init__(*args, **kwargs)


SonoraSpectrum = Sonora2021Spectrum


class SonoraGrid(SpectrumCollection):
    r"""
    A container for a grid of Sonora precomputed synthetic spectra of brown dwarfs and free-floating
    Gas Giant planets.

    Args:
        Teff_range (tuple): The Teff limits of the grid model to read in.
        logg (tuple): The logg limits of the Sonora model to read in.
        metallicity_range (tuple): The metallicity limits of the Sonora model to read in
        path (str): The path to your local Sonora grid library.
            You must have the Sonora grid downloaded locally.
            Default: "~/libraries/raw/Sonora/"
        wl_lo (float): the bluest wavelength of the models to keep (Angstroms)
        wl_hi (float): the reddest wavelength of the models to keep (Angstroms)
        """

    def __init__(
        self,
        teff_range=None,
        logg_range=None,
        metallicity_range=None,
        path=None,
        wl_lo=8038,
        wl_hi=12849,
        **kwargs,
    ):

        if (
            ("flux" in kwargs.keys())
            and ("spectral_axis" in kwargs.keys())
            and ("meta" in kwargs.keys())
        ):
            super().__init__(**kwargs)
        else:
            teff_points = np.hstack(
                (
                    np.arange(500, 600, 25),
                    np.arange(600, 1000, 50),
                    np.arange(1000, 2401, 100),
                )
            )
            logg_points = np.arange(3.0, 5.51, 0.25)

            metallicity_points = np.arange(-0.5, 0.51, 0.5)

            if teff_range is not None:
                subset = (teff_points >= teff_range[0]) & (teff_points <= teff_range[1])
                teff_points = teff_points[subset]

            if logg_range is not None:
                subset = (logg_points >= logg_range[0]) & (logg_points <= logg_range[1])
                logg_points = logg_points[subset]

            if metallicity_range is not None:
                subset = (metallicity_points >= metallicity_range[0]) & (metallicity_points <= metallicity_range[1])
                metallicity_points = metallicity_points[subset]

            wavelengths, fluxes = [], []
            grid_points = []

            pbar = tqdm(teff_points)
            for teff in pbar:
                for logg in logg_points:
                    for metallicity in metallicity_points:
                        pbar.set_description(
                            "Processing Teff={} K, logg={:0.2f}, metallicity={:0.1f}".format(
                                teff, logg, metallicity
                            )
                        )
                        grid_point = (teff, logg, metallicity)
                        # "To do": See issue 31
                        # Temporary work around: pass if we can't read the file
                        try:
                            spec = SonoraSpectrum(
                                teff=teff,
                                logg=logg,
                                metallicity=metallicity,
                                path=path,
                                wl_lo=wl_lo,
                                wl_hi=wl_hi,
                            )
                            wavelengths.append(spec.wavelength)
                            fluxes.append(spec.flux)
                            grid_points.append(grid_point)
                        except:
                            log.info(
                                "Grid point Teff={} K, logg={:0.2f}, metallicity={:0.1f} does not exist".format(
                                    teff, logg, metallicity
                                )
                                + " in Sonora library"
                            )
            flux_out = np.array(fluxes) * fluxes[0].unit
            wave_out = np.array(wavelengths) * wavelengths[0].unit

            # Make a quick-access dictionary
            n_spectra = len(grid_points)
            lookup_dict = {grid_points[i]: i for i in range(n_spectra)}
            meta = {
                "teff_points": teff_points,
                "logg_points": logg_points,
                "metallicity_points": metallicity_points,
                "grid_labels": ("T_eff", "log(g)"),
                "n_spectra": n_spectra,
                "grid_points": grid_points,
                "lookup_dict": lookup_dict,
            }

            super().__init__(flux=flux_out, spectral_axis=wave_out, meta=meta)

    def __getitem__(self, key):
        flux = self.flux[key]
        if flux.ndim != 1:
            raise ValueError(
                "Currently only 1D data structures may be "
                "returned from slice operations."
            )
        spectral_axis = self.spectral_axis[key]
        uncertainty = None if self.uncertainty is None else self.uncertainty[key]
        mask = None if self.mask is None else self.mask[key]
        if self.meta is None:
            meta = None
        else:
            try:
                meta = self.meta[key]
            except KeyError:
                meta = self.meta

        return SonoraSpectrum(
            flux=flux,
            spectral_axis=spectral_axis,
            uncertainty=uncertainty,
            wcs=None,
            mask=mask,
            meta=meta,
        )

    @property
    def grid_points(self):
        """What are the coordinates of the grid?"""
        return self.meta["grid_points"]

    @property
    def teff_points(self):
        """What are the Teff points of the grid?"""
        return self.meta["teff_points"]

    @property
    def logg_points(self):
        """What are the logg points of the grid?"""
        return self.meta["logg_points"]

    @property
    def metallicity_points(self):
        """What are the metallicity points of the grid?"""
        return self.meta["metallicity_points"]

    @property
    def grid_labels(self):
        """What are the grid labels?"""
        return self.meta["grid_labels"]

    @property
    def n_spectra(self):
        """How many distinct spectra are in the grid?"""
        return self.meta["n_spectra"]

    @property
    def lookup_dict(self):
        """Lookup dictioary for spectra from their grid coordinates"""
        return self.meta["lookup_dict"]

    def truncate(self, wavelength_range=None, data=None):
        """Truncate the wavelength range of the grid

        Parameters
        ----------
        wavelength_range: List or Tuple of Quantities
            A pair of values that denote the shortest and longest wavelengths
            for truncating the grid.
        data: Spectrum1D-like
            A spectrum to which this method will match the wavelength limits

        """
        fiducial_spectrum = copy.deepcopy(self[0])
        wavelength_units = fiducial_spectrum.wavelength.unit
        flux_units = fiducial_spectrum.flux.unit

        if (data is not None) and (wavelength_range is None):
            wavelength_range = (
                fiducial_spectrum.wavelength.value.min() * wavelength_units,
                fiducial_spectrum.wavelength.value.max() * wavelength_units,
            )
        shortest_wavelength, longest_wavelength = wavelength_range

        wavelengths, fluxes = [], []
        for spectrum in self:
            mask = (spectrum.wavelength > shortest_wavelength) & (
                spectrum.wavelength < longest_wavelength
            )
            wavelengths.append(spectrum.wavelength.value[mask])
            fluxes.append(spectrum.flux.value[mask])

        fluxes = np.array(fluxes) * flux_units
        wavelengths = np.array(wavelengths) * wavelength_units
        assert fluxes is not None
        assert wavelengths is not None

        return self.__class__(flux=fluxes, spectral_axis=wavelengths, meta=self.meta)

    def get_index(self, grid_point):
        """Get the spectrum index associated with a given grid point
        """
        return self.lookup_dict[grid_point]

    def get_distance(self, gridpoint1, gridpoint2):
        return sqrt(((gridpoint1[0]-gridpoint2[0]) ** 2) + ((gridpoint1[1]-gridpoint2[1]) ** 2) + ((gridpoint1[2]-gridpoint2[2]) ** 2))

    #  Need to add a function to find the near grid point in the case it doesn't exist (find nearest point in a lattice)
    def find_nearest_grid_point(self, teff, logg, metallicity):
        current = (teff, logg, metallicity)

        distances = []
        for point in self.grid_points:
            distances.append(self.get_distance(current, point))
        shortest_distance = min(distances)
        idx = distances.index(shortest_distance)
        nearest_point = self.grid_points[idx]

        return nearest_point

    def find_nearest_teff(self, value):
        idx = (np.abs(self.teff_points - value)).argmin()
        return self.teff_points[idx]

    def show_dashboard(
        self, data=None, notebook_url="localhost:8888", show_telluric=True
    ):
        """Show an interactive dashboard for interacting with the Sonora grid
        Heavily inspired by the lightkurve .interact() method.

        Parameters
        ----------
        data: Spectrum1D-like
            A normalized data spectrum over which to plot the models
        notebook_url: str
            Location of the Jupyter notebook page (default: "localhost:8888")
            When showing Bokeh applications, the Bokeh server must be
            explicitly configured to allow connections originating from
            different URLs. This parameter defaults to the standard notebook
            host and port. If you are running on a different location, you
            will need to supply this value for the application to display
            properly. If no protocol is supplied in the URL, e.g. if it is
            of the form "localhost:8888", then "http" will be used.

        """

        def create_interact_ui(doc):

            # Make the spectrum source
            scalar_norm = np.percentile(self[0].flux.value, 95)
            spec_source = ColumnDataSource(
                data=dict(
                    wavelength=self[0].wavelength,
                    flux=self[0].flux.value / scalar_norm,
                    native_flux=self[0].flux.value / scalar_norm,
                    native_wavelength=self[0].wavelength.value,
                )
            )

            fig = figure(
                title="Sonora Bobcat Interactive Dashboard",
                plot_height=340,
                plot_width=600,
                tools="pan,wheel_zoom,box_zoom,reset,save",
                toolbar_location="below",
                border_fill_color="whitesmoke",
            )
            fig.title.offset = -10
            fig.yaxis.axis_label = "Flux "
            fig.xaxis.axis_label = "Wavelength (Angstrom)"
            fig.y_range = Range1d(start=0, end=1.9)

            wl_lo, wl_hi = (
                self[0].wavelength.value.min(),
                self[0].wavelength.value.max(),
            )

            instrumental_resolution = 100_000
            if data is not None:
                assert isinstance(
                    data, Spectrum1D
                ), "The data spectrum must be Spectrum1D-like"
                new_lo, new_hi = (
                    data.wavelength.value.min(),
                    data.wavelength.value.max(),
                )
                assert (new_lo < wl_hi) & (
                    new_hi > wl_lo
                ), "Data should overlap the models, double check your wavelength limits."
                wl_lo, wl_hi = new_lo, new_hi

                data_source = ColumnDataSource(
                    data=dict(wavelength=data.wavelength.value, flux=data.flux.value,)
                )
                fig.step(
                    "wavelength",
                    "flux",
                    line_width=2,
                    legend_label="data",
                    source=data_source,
                )
                if hasattr(data, "instrumental_resolution"):
                    instrumental_resolution = data.instrumental_resolution

            fig.x_range = Range1d(start=wl_lo, end=wl_hi)

            if show_telluric:
                tell_spec = TelluricSpectrum(
                    path="default", wl_lo=wl_lo, wl_hi=wl_hi
                ).instrumental_broaden(instrumental_resolution * 2)
                tell_source = ColumnDataSource(
                    data=dict(
                        wavelength=tell_spec.wavelength.value,
                        flux=tell_spec.flux.value,
                    )
                )
                out_glyph = fig.step(
                    "wavelength",
                    "flux",
                    line_width=2,
                    color="#bdc3c7",
                    legend_label="Telluric",
                    source=tell_source,
                )
                out_glyph.level = "underlay"

            fig.step(
                "wavelength",
                "flux",
                line_width=2,
                color="Red",
                source=spec_source,
                legend_label="Sonora Model",
                nonselection_line_color="DarkOrange",
                nonselection_line_alpha=1.0,
            )

            fig.legend.location = "top_left"
            fig.legend.orientation = "horizontal"

            # Slider to decimate the data
            smoothing_slider = Slider(
                start=0.1,
                end=40,
                value=0.1,
                step=0.1,
                title="Rotational Broadening: v sin(i) [km/s]",
                width=490,
            )

            vz_slider = Slider(
                start=-200,
                end=200,
                value=0.00,
                step=0.05,
                title="Radial Velocity: RV [km/s]",
                width=490,
                format="0.000f",
            )

            teff_slider = Slider(
                start=min(self.teff_points),
                end=max(self.teff_points),
                value=1000,
                step=25,
                title="Effective Temperature: T_eff [Kelvin]",
                width=490,
            )
            teff_message = Div(
                text="Closest T_eff point: {}".format(1000), width=100, height=10
            )
            logg_slider = Slider(
                start=min(self.logg_points),
                end=max(self.logg_points),
                value=5.0,
                step=0.25,
                title="Surface Gravity: log(g) [cm/s^2]",
                width=490,
            )
            logg_message = Div(
                text="Closest log(g) point: {}".format(1000), width=100, height=10
            )
            metallicity_slider = Slider(
                start=min(self.metallicity_points),
                end=max(self.metallicity_points),
                value=0.0,
                step=0.5,
                title="Metallicity: Metallicity [Fe/H]",
                width=490,
            )
            metallicity_message = Div(
                text="Closest Metallicity point: {}".format(1000), width=100, height=10
            )
            scale_slider = Slider(
                start=0.1,
                end=2.0,
                value=1.0,
                step=0.005,
                title="Normalization Scalar",
                width=490,
            )
            r_button = Button(label=">", button_type="default", width=30)
            l_button = Button(label="<", button_type="default", width=30)

            def update_upon_scale(attr, old, new):
                """Callback to take action when smoothing slider changes"""
                new_spec = (
                    SonoraSpectrum(
                        spectral_axis=spec_source.data["native_wavelength"]
                        * u.Angstrom,
                        flux=spec_source.data["native_flux"] * u.dimensionless_unscaled,
                    )
                    .rotationally_broaden(smoothing_slider.value)
                    .multiply(new * u.dimensionless_unscaled)
                    .rv_shift(vz_slider.value)
                )
                spec_source.data["flux"] = new_spec.flux.value

            def update_upon_smooth(attr, old, new):
                """Callback to take action when smoothing slider changes"""
                new_spec = (
                    SonoraSpectrum(
                        spectral_axis=spec_source.data["native_wavelength"]
                        * u.Angstrom,
                        flux=spec_source.data["native_flux"] * u.dimensionless_unscaled,
                    )
                    .rotationally_broaden(new)
                    .multiply(scale_slider.value * u.dimensionless_unscaled)
                    .rv_shift(vz_slider.value)
                )
                spec_source.data["flux"] = new_spec.flux.value

            def update_upon_vz(attr, old, new):
                """Callback to take action when vz slider changes"""
                new_spec = SonoraSpectrum(
                    spectral_axis=spec_source.data["native_wavelength"] * u.Angstrom,
                    flux=spec_source.data["native_flux"] * u.dimensionless_unscaled,
                ).rv_shift(new)
                spec_source.data["wavelength"] = new_spec.wavelength.value

            def update_upon_teff_selection(attr, old, new):
                """Callback to take action when teff slider changes"""
                logg = logg_slider.value
                metallicity = metallicity_slider.value
                new_grid_point = self.find_nearest_grid_point(new, logg, metallicity)

                if new_grid_point[0] != old:
                    teff = new_grid_point[0]
                    logg = new_grid_point[1]
                    metallicity = new_grid_point[2]

                    teff_message.text = "Closest T_eff point: {}".format(teff)
                    logg_message.text = "Closest log(g) point: {}".format(logg)
                    metallicity_message.text = "Closest Metallicity point: {}".format(metallicity)
                    index = self.get_index(new_grid_point)

                    native_spec = self[index].normalize(percentile=95)
                    new_spec = (
                        native_spec.rotationally_broaden(smoothing_slider.value)
                        .multiply(scale_slider.value * u.dimensionless_unscaled)
                        .rv_shift(vz_slider.value)
                    )

                    spec_source.data = {
                        "native_wavelength": native_spec.wavelength.value,
                        "native_flux": native_spec.flux.value,
                        "wavelength": new_spec.wavelength.value,
                        "flux": new_spec.flux.value,
                    }
                else:
                    pass

            def update_upon_logg_selection(attr, old, new):
                """Callback to take action when logg slider changes"""
                teff = teff_slider.value
                metallicity = metallicity_slider.value
                new_grid_point = self.find_nearest_grid_point(teff, new, metallicity)

                if new_grid_point[1] != old:
                    teff = new_grid_point[0]
                    logg = new_grid_point[1]
                    metallicity = new_grid_point[2]

                    teff_message.text = "Closest T_eff point: {}".format(teff)
                    logg_message.text = "Closest log(g) point: {}".format(logg)
                    metallicity_message.text = "Closest Metallicity point: {}".format(metallicity)
                    index = self.get_index(new_grid_point)

                    native_spec = self[index].normalize(percentile=95)
                    new_spec = (
                        native_spec.rotationally_broaden(smoothing_slider.value)
                        .multiply(scale_slider.value * u.dimensionless_unscaled)
                        .rv_shift(vz_slider.value)
                    )

                    spec_source.data = {
                        "native_wavelength": native_spec.wavelength.value,
                        "native_flux": native_spec.flux.value,
                        "wavelength": new_spec.wavelength.value,
                        "flux": new_spec.flux.value,
                    }

                else:
                    pass
            def update_upon_metallicity_selection(attr, old, new):
                """Callback to take action when metallicity slider changes"""
                teff = teff_slider.value
                logg = logg_slider.value
                new_grid_point = self.find_nearest_grid_point(teff, logg, new)

                if new_grid_point[2] != old:
                    teff = new_grid_point[0]
                    logg = new_grid_point[1]
                    metallicity = new_grid_point[2]

                    teff_message.text = "Closest T_eff point: {}".format(teff)
                    logg_message.text = "Closest log(g) point: {}".format(logg)
                    metallicity_message.text = "Closest Metallicity point: {}".format(metallicity)
                    index = self.get_index(new_grid_point)

                    native_spec = self[index].normalize(percentile=95)
                    new_spec = (
                        native_spec.rotationally_broaden(smoothing_slider.value)
                        .multiply(scale_slider.value * u.dimensionless_unscaled)
                        .rv_shift(vz_slider.value)
                    )

                    spec_source.data = {
                        "native_wavelength": native_spec.wavelength.value,
                        "native_flux": native_spec.flux.value,
                        "wavelength": new_spec.wavelength.value,
                        "flux": new_spec.flux.value,
                    }
                else:
                    pass

            def go_right_by_one():
                """Step forward in time by a single cadence"""
                current_index = np.abs(self.teff_points - teff_slider.value).argmin()
                new_index = current_index + 1
                if new_index <= (len(self.teff_points) - 1):
                    teff_slider.value = self.teff_points[new_index]

            def go_left_by_one():
                """Step back in time by a single cadence"""
                current_index = np.abs(self.teff_points - teff_slider.value).argmin()
                new_index = current_index - 1
                if new_index >= 0:
                    teff_slider.value = self.teff_points[new_index]

            r_button.on_click(go_right_by_one)
            l_button.on_click(go_left_by_one)
            smoothing_slider.on_change("value", update_upon_smooth)
            vz_slider.on_change("value", update_upon_vz)
            teff_slider.on_change("value", update_upon_teff_selection)
            logg_slider.on_change("value", update_upon_logg_selection)
            metallicity_slider.on_change("value", update_upon_metallicity_selection)
            scale_slider.on_change("value", update_upon_scale)

            sp1, sp2, sp3, sp4 = (
                Spacer(width=5),
                Spacer(width=10),
                Spacer(width=20),
                Spacer(width=100),
            )

            widgets_and_figures = layout(
                [fig],
                [l_button, sp1, r_button, sp2, teff_slider, sp3, teff_message],
                [sp4, logg_slider, sp3, logg_message],
                [sp4, metallicity_slider, sp3, metallicity_message], 
                [sp4, smoothing_slider],
                [sp4, vz_slider],
                [sp4, scale_slider],
            )
            doc.add_root(widgets_and_figures)

        output_notebook(verbose=False, hide_banner=True)
        show(create_interact_ui, notebook_url=notebook_url)



r"""
Telluric Spectrum
-----------------

An experimental container for a single grid-point of a pre-computed telluric atmosphere model.  The spectrum is a vector with coordinates wavelength and flux :math:`F(\lambda)`.

TelluricSpectrum
################
"""

import warnings
import logging
from gollum.precomputed_spectrum import PrecomputedSpectrum
import numpy as np
import astropy
import pandas as pd
from astropy import units as u
from specutils import SpectrumCollection
import os
from specutils.spectra.spectrum1d import Spectrum1D
from tqdm import tqdm

log = logging.getLogger(__name__)

#  See Issue: https://github.com/astropy/specutils/issues/779
warnings.filterwarnings(
    "ignore", category=astropy.utils.exceptions.AstropyDeprecationWarning
)
# See Issue: https://github.com/astropy/specutils/issues/800
warnings.filterwarnings("ignore", category=RuntimeWarning)


## TODO: rename this to SkyCalc??
class TelluricSpectrum(PrecomputedSpectrum):
    r"""
    A container for a single Telluric precomputed synthetic spectrum, currently from the skycalc website at eso.org. 

    Args:
        path (str): The path to a whitespace-delimited ASCII file generated by skycalc on eso.org.  You can either provide the file locally or a weblink.  'default' uses a demo spectrum
        wl_lo (float): the bluest wavelength of the models to keep (Angstroms)
        wl_hi (float): the reddest wavelength of the models to keep (Angstroms)
        """

    def __init__(self, *args, path=None, wl_lo=8038, wl_hi=12849, **kwargs):

        if path == "default":
            path = "https://gist.githubusercontent.com/gully/b93a04ede17f617d36b2a8ffc32a60e6/raw/01c603ee201509b79b1a464093e6200ae2032198/demo_skycalc_telluric_transmission_spectrum.txt"

        if path is not None:

            # Units: nm, transmittance
            df_native = (
                pd.read_csv(
                    path,
                    delim_whitespace=True,
                    names=["wavelength_nm", "transmittance"],
                )
                .sort_values("wavelength_nm")
                .reset_index(drop=True)
            )

            # convert to Angstrom
            df_native["wavelength"] = df_native["wavelength_nm"] * 10.0
            mask = (df_native.wavelength > wl_lo) & (df_native.wavelength < wl_hi)
            df_trimmed = df_native[mask].reset_index(drop=True)

            super().__init__(
                spectral_axis=df_trimmed.wavelength.values * u.Angstrom,
                flux=df_trimmed.transmittance.values * u.dimensionless_unscaled,
                **kwargs,
            )

        else:
            super().__init__(*args, **kwargs)


class TelFitSpectrum(PrecomputedSpectrum):
    r"""
    A container for a single TelFit precomputed synthetic spectrum. 

    Args:
        path (str): The path to a whitespace-delimited ASCII file generated by skycalc on eso.org.  You can either provide the file locally or a weblink.  'default' uses a demo spectrum
        wl_lo (float): the bluest wavelength of the models to keep (Angstroms)
        wl_hi (float): the reddest wavelength of the models to keep (Angstroms)
        """

    def __init__(self, *args, path=None, **kwargs):

        if path is not None:

            names = ["wavelength_nm", "transmission", "continuum", "err"]
            df_native = pd.read_csv(
                path,
                delim_whitespace=True,
                names=names,
                usecols=names[0:2],
                encoding="unicode_escape",  # or UTF-8?
                low_memory=False,
            )

            # Units: nm, transmittance
            # convert to Angstrom
            df_native["wavelength"] = (
                pd.to_numeric(df_native["wavelength_nm"], errors="coerce") * 10.0
            )
            flux_out = pd.to_numeric(df_native["transmission"], errors="coerce")

            good_mask = (df_native.wavelength.values == df_native.wavelength.values) & (
                flux_out.values == flux_out.values
            )

            super().__init__(
                spectral_axis=df_native.wavelength.values[good_mask] * u.Angstrom,
                flux=flux_out.values[good_mask] * u.dimensionless_unscaled,
                **kwargs,
            )

        else:
            super().__init__(*args, **kwargs)

    def air_to_vacuum(self):
        """Converts a spectrum from air to vacuum  

        Based on Morton, D. C. 1991, ApJS, 77, 119
        """
        wave_A = self.wavelength.to(u.Angstrom).value
        n_air = 1.0 + 2.735182e-4 + 131.4182 / wave_A ** 2 + 2.76249e8 / wave_A ** 4
        vacuum_wavelengths = self.wavelength * n_air
        return self._copy(spectral_axis=vacuum_wavelengths, wcs=None)




r"""
PHOENIX Spectrum
-----------------

A container for a single Phoenix grid-point spectrum of wavelength and flux :math:`F(\lambda)`.

PHOENIXSpectrum
###############
"""

import copy
import warnings
import logging
from gollum.precomputed_spectrum import PrecomputedSpectrum
import numpy as np
import astropy
from astropy.io import fits
from astropy import units as u
from specutils import SpectrumCollection
from specutils.spectra.spectrum1d import Spectrum1D
from tqdm import tqdm
import os

from bokeh.io import show, output_notebook, push_notebook
from bokeh.plotting import figure, ColumnDataSource
from bokeh.models import Slider, Span, Range1d, Dropdown
from bokeh.layouts import layout, Spacer
from bokeh.models.widgets import Button, Div

log = logging.getLogger(__name__)

#  See Issue: https://github.com/astropy/specutils/issues/779
warnings.filterwarnings(
    "ignore", category=astropy.utils.exceptions.AstropyDeprecationWarning
)
# See Issue: https://github.com/astropy/specutils/issues/800
warnings.filterwarnings("ignore", category=RuntimeWarning)


class PHOENIXSpectrum(PrecomputedSpectrum):
    r"""
    A container for PHOENIX spectra

    Args:
        Teff (int): The Teff label of the PHOENIX model to read in.  Must be on the PHOENIX grid.
        logg (float): The logg label of the PHOENIX model to read in.  Must be on the PHOENIX grid.
        path (str): The path to your local PHOENIX grid library.  You must have the PHOENIX
            grid downloaded locally.  Default: "~/libraries/raw/PHOENIX/"
        download (bool): **Experimental** Whether or not you want to download the spectra 
            from the internet.  Requires an internet connection to work.
        wl_lo (float): the bluest wavelength of the models to keep (Angstroms)
        wl_hi (float): the reddest wavelength of the models to keep (Angstroms)
    """

    def __init__(
        self,
        *args,
        teff=None,
        logg=None,
        metallicity=None,
        path=None,
        download=False,
        wl_lo=8038,
        wl_hi=12849,
        **kwargs,
    ):

        if metallicity is None:
            metallicity = 0.0  # solar by default

        if path is None:
            path = "~/libraries/raw/PHOENIX/"

        if download == False:
            base_path = os.path.expanduser(path)
            assert os.path.exists(
                base_path
            ), "You must specify the path to local PHOENIX models"

            wl_filename = base_path + "/WAVE_PHOENIX-ACES-AGSS-COND-2011.fits"
            assert os.path.exists(
                wl_filename
            ), f"You need to place the PHOENIX models in {base_path}"
        else:
            log.info(
                "Experimental feature! Attempting to download PHOENIX models from the internet..."
            )
            log.info(
                "We are using this FTP site: ftp://phoenix.astro.physik.uni-goettingen.de/v2.0/HiResFITS/"
            )
            wl_filename = "ftp://phoenix.astro.physik.uni-goettingen.de/v2.0/HiResFITS/WAVE_PHOENIX-ACES-AGSS-COND-2011.fits"
            base_path = "ftp://phoenix.astro.physik.uni-goettingen.de/v2.0/HiResFITS/PHOENIX-ACES-AGSS-COND-2011/"

        if (teff is not None) & (logg is not None):

            wl_orig = fits.open(wl_filename)[0].data.astype(np.float64)

            mask = (wl_orig > wl_lo) & (wl_orig < wl_hi)
            wl_out = wl_orig[mask]

            # Deal with metallicity
            metallicity_string = f"{metallicity:+0.1f}"

            metallicity_string = "-0.0" if metallicity == 0.0 else metallicity_string

            fn = (
                base_path
                + "/Z{}/lte{:05d}-{:0.2f}{}.PHOENIX-ACES-AGSS-COND-2011-HiRes.fits"
            ).format(metallicity_string, teff, logg, metallicity_string)

            flux_orig = fits.open(fn)[0].data.astype(np.float64)
            # Units: erg/s/cm^2/cm
            flux_native = flux_orig[mask]

            native_flux_units = u.erg / u.s / u.cm ** 2 / u.cm

            meta_dict = {
                "teff": teff,
                "logg": logg,
                "metallicity": metallicity,
                "native_flux_unit": native_flux_units,
            }

            super().__init__(
                spectral_axis=wl_out * u.AA,
                flux=flux_native * native_flux_units,
                meta=meta_dict,
                **kwargs,
            )

        else:
            super().__init__(*args, **kwargs)

    @property
    def teff(self):
        """The input Effective Temperature associated with this model"""
        if "teff" in self.meta.keys():
            return self.meta["teff"]
        else:
            return None

    @property
    def logg(self):
        """The input surface gravity associated with this model"""
        if "logg" in self.meta.keys():
            return self.meta["logg"]
        else:
            return None

    @property
    def metallicity(self):
        """The input metallicity associated with this model"""
        if "metallicity" in self.meta.keys():
            return self.meta["metallicity"]
        else:
            return None


class PHOENIXGrid(SpectrumCollection):
    r"""
    A container for a grid of PHOENIX precomputed synthetic spectra of stars.

    Args:
        Teff_range (tuple): The Teff limits of the grid model to read in.
        logg (tuple): The logg limits of the grid model to read in.
        path (str): The path to your local PHOENIX grid library.
            You must have the PHOENIX grid downloaded locally.
            Default: "~/libraries/raw/PHOENIX/"
        wl_lo (float): the bluest wavelength of the models to keep (Angstroms)
        wl_hi (float): the reddest wavelength of the models to keep (Angstroms)
    """

    def __init__(
        self,
        teff_range=None,
        logg_range=None,
        metallicity_range=None,
        path=None,
        wl_lo=8038,
        wl_hi=12849,
        **kwargs,
    ):

        if (
            ("flux" in kwargs.keys())
            and ("spectral_axis" in kwargs.keys())
            and ("meta" in kwargs.keys())
        ):
            # Trigger a passthrough
            super().__init__(**kwargs)
        else:

            teff_points = np.hstack(
                (np.arange(2300, 7000, 100), np.arange(7000, 12_001, 200))
            )
            # Todo: some T_eff ranges go to log(g) = 0.0, consider adding these
            logg_points = np.arange(2.0, 6.01, 0.5)

            metallicity_points = np.array([-4, -3, -2, -1.5, -1, -0.5, 0, 0.5, 1])

            if teff_range is not None:
                subset = (teff_points >= teff_range[0]) & (teff_points <= teff_range[1])
                teff_points = teff_points[subset]

            if logg_range is not None:
                subset = (logg_points >= logg_range[0]) & (logg_points <= logg_range[1])
                logg_points = logg_points[subset]

            if metallicity_range is not None:
                subset = (metallicity_points >= metallicity_range[0]) & (
                    metallicity_points <= metallicity_range[1]
                )
                metallicity_points = metallicity_points[subset]

            wavelengths, fluxes = [], []
            grid_points = []

            pbar = tqdm(teff_points)
            for teff in pbar:
                for logg in logg_points:
                    for metallicity in metallicity_points:
                        pbar.set_description(
                            "Processing Teff={} K, logg={:0.2f}, Z={:+0.1f}".format(
                                teff, logg, metallicity
                            )
                        )
                        grid_point = (teff, logg, metallicity)
                        spec = PHOENIXSpectrum(
                            teff=teff,
                            logg=logg,
                            metallicity=metallicity,
                            path=path,
                            wl_lo=wl_lo,
                            wl_hi=wl_hi,
                        )
                        wavelengths.append(spec.wavelength)
                        fluxes.append(spec.flux)
                        grid_points.append(grid_point)
            flux_out = np.array(fluxes) * fluxes[0].unit
            wave_out = np.array(wavelengths) * wavelengths[0].unit

            # Make a quick-access dictionary
            n_spectra = len(grid_points)
            lookup_dict = {grid_points[i]: i for i in range(n_spectra)}
            meta = {
                "teff_points": teff_points,
                "logg_points": logg_points,
                "metallicity_points": metallicity_points,
                "grid_labels": ("T_eff", "log(g)", "Z"),
                "n_spectra": n_spectra,
                "grid_points": grid_points,
                "lookup_dict": lookup_dict,
            }

            super().__init__(flux=flux_out, spectral_axis=wave_out, meta=meta)

    def __getitem__(self, key):
        flux = self.flux[key]
        if flux.ndim != 1:
            raise ValueError(
                "Currently only 1D data structures may be "
                "returned from slice operations."
            )
        spectral_axis = self.spectral_axis[key]
        uncertainty = None if self.uncertainty is None else self.uncertainty[key]
        wcs = None if self.wcs is None else self.wcs[key]
        mask = None if self.mask is None else self.mask[key]
        if self.meta is None:
            meta = None
        else:
            try:
                meta = self.meta[key]
            except KeyError:
                meta = self.meta

        return PHOENIXSpectrum(
            flux=flux,
            spectral_axis=spectral_axis,
            uncertainty=uncertainty,
            wcs=wcs,
            mask=mask,
            meta=meta,
        )

    @property
    def grid_points(self):
        """What are the coordinates of the grid?"""
        return self.meta["grid_points"]

    @property
    def teff_points(self):
        """What are the Teff points of the grid?"""
        return self.meta["teff_points"]

    @property
    def metallicity_points(self):
        """What are the metallicity points of the grid?"""
        return self.meta["metallicity_points"]

    @property
    def logg_points(self):
        """What are the logg points of the grid?"""
        return self.meta["logg_points"]

    @property
    def grid_labels(self):
        """What are the grid labels?"""
        return self.meta["grid_labels"]

    @property
    def n_spectra(self):
        """How many distinct spectra are in the grid?"""
        return self.meta["n_spectra"]

    @property
    def lookup_dict(self):
        """Lookup dictioary for spectra from their grid coordinates"""
        return self.meta["lookup_dict"]

    def truncate(self, wavelength_range=None, data=None):
        """Truncate the wavelength range of the grid

        Parameters
        ----------
        wavelength_range: List or Tuple of Quantities
            A pair of values that denote the shortest and longest wavelengths
            for truncating the grid.
        data: Spectrum1D-like
            A spectrum to which this method will match the wavelength limits

        """
        fiducial_spectrum = copy.deepcopy(self[0])
        wavelength_units = fiducial_spectrum.wavelength.unit
        flux_units = fiducial_spectrum.flux.unit

        if (data is not None) and (wavelength_range is None):
            wavelength_range = (
                fiducial_spectrum.wavelength.value.min() * wavelength_units,
                fiducial_spectrum.wavelength.value.max() * wavelength_units,
            )
        shortest_wavelength, longest_wavelength = wavelength_range

        wavelengths, fluxes = [], []
        for spectrum in self:
            mask = (spectrum.wavelength > shortest_wavelength) & (
                spectrum.wavelength < longest_wavelength
            )
            wavelengths.append(spectrum.wavelength.value[mask])
            fluxes.append(spectrum.flux.value[mask])

        fluxes = np.array(fluxes) * flux_units
        wavelengths = np.array(wavelengths) * wavelength_units
        assert fluxes is not None
        assert wavelengths is not None

        return self.__class__(flux=fluxes, spectral_axis=wavelengths, meta=self.meta)

    def get_index(self, grid_point):
        """Get the spectrum index associated with a given grid point"""
        return self.lookup_dict[grid_point]

    def find_nearest_teff(self, value):
        idx = (np.abs(self.teff_points - value)).argmin()
        return self.teff_points[idx]

    def find_nearest_metallicity(self, value):
        idx = (np.abs(self.metallicity_points - value)).argmin()
        return self.metallicity_points[idx]

    def show_dashboard(self, data=None, notebook_url="localhost:8888"):
        """Show an interactive dashboard for interacting with the PHOENIX grid
        Heavily inspired by the lightkurve .interact() method.

        Parameters
        ----------
        data: Spectrum1D-like
            A normalized data spectrum over which to plot the models
        notebook_url: str
            Location of the Jupyter notebook page (default: "localhost:8888")
            When showing Bokeh applications, the Bokeh server must be
            explicitly configured to allow connections originating from
            different URLs. This parameter defaults to the standard notebook
            host and port. If you are running on a different location, you
            will need to supply this value for the application to display
            properly. If no protocol is supplied in the URL, e.g. if it is
            of the form "localhost:8888", then "http" will be used.

        """

        def create_interact_ui(doc):

            # Make the spectrum source
            scalar_norm = np.percentile(self[0].flux.value, 95)
            spec_source = ColumnDataSource(
                data=dict(
                    wavelength=self[0].wavelength.value,
                    flux=self[0].flux.value / scalar_norm,
                    native_flux=self[0].flux.value / scalar_norm,
                    native_wavelength=self[0].wavelength.value,
                )
            )

            fig = figure(
                title="PHOENIX Interactive Dashboard",
                plot_height=340,
                plot_width=600,
                tools="pan,wheel_zoom,box_zoom,tap,reset",
                toolbar_location="below",
                border_fill_color="whitesmoke",
            )
            fig.title.offset = -10
            fig.yaxis.axis_label = "Flux "
            fig.xaxis.axis_label = "Wavelength (micron)"
            fig.y_range = Range1d(start=0, end=1.5)

            fig.step(
                "wavelength",
                "flux",
                line_width=1,
                color="gray",
                source=spec_source,
                nonselection_line_color="gray",
                nonselection_line_alpha=1.0,
            )
            wl_lo, wl_hi = (
                self[0].wavelength.value.min(),
                self[0].wavelength.value.max(),
            )

            if data is not None:
                assert isinstance(
                    data, Spectrum1D
                ), "The data spectrum must be Spectrum1D-like"
                new_lo, new_hi = (
                    data.wavelength.value.min(),
                    data.wavelength.value.max(),
                )
                assert (new_lo < wl_hi) & (
                    new_hi > wl_lo
                ), "Data should overlap the models, double check your wavelength limits."
                wl_lo, wl_hi = new_lo, new_hi

                data_source = ColumnDataSource(
                    data=dict(wavelength=data.wavelength.value, flux=data.flux.value,)
                )
                fig.step(
                    "wavelength", "flux", line_width=1, color="blue", source=data_source
                )

            fig.x_range = Range1d(start=wl_lo, end=wl_hi)

            # Slider to decimate the data
            smoothing_slider = Slider(
                start=0.1,
                end=200,
                value=0.1,
                step=0.1,
                title="Rotational Broadening: v sin(i) [km/s]",
                width=490,
            )

            vz_slider = Slider(
                start=-200,
                end=200,
                value=0.00,
                step=0.05,
                title="Radial Velocity: RV [km/s]",
                width=490,
                format="0.000f",
            )

            teff_slider = Slider(
                start=min(self.teff_points),
                end=max(self.teff_points),
                value=self.teff_points[1],
                step=100,
                title="Effective Temperature: T_eff [Kelvin]",
                width=490,
            )
            teff_message = Div(
                text="Closest grid point: {}".format(1000), width=100, height=10
            )
            logg_slider = Slider(
                start=min(self.logg_points),
                end=max(self.logg_points),
                value=5.0,
                step=0.50,
                title="Surface Gravity: log(g) [cm/s^2]",
                width=490,
            )

            metallicity_slider = Slider(
                start=min(self.metallicity_points),
                end=max(self.metallicity_points),
                value=0.0,
                step=0.50,
                title="Metallicity: Z",
                width=490,
            )

            r_button = Button(label=">", button_type="default", width=30)
            l_button = Button(label="<", button_type="default", width=30)

            def update_upon_smooth(attr, old, new):
                """Callback to take action when smoothing slider changes"""
                new_spec = PHOENIXSpectrum(
                    spectral_axis=spec_source.data["native_wavelength"] * u.Angstrom,
                    flux=spec_source.data["native_flux"] * u.dimensionless_unscaled,
                ).rotationally_broaden(new)
                spec_source.data["flux"] = new_spec.flux.value

            def update_upon_vz(attr, old, new):
                """Callback to take action when vz slider changes"""
                new_spec = PHOENIXSpectrum(
                    spectral_axis=spec_source.data["native_wavelength"] * u.Angstrom,
                    flux=spec_source.data["native_flux"] * u.dimensionless_unscaled,
                ).rv_shift(new)
                spec_source.data["wavelength"] = new_spec.wavelength.value

            def update_upon_teff_selection(attr, old, new):
                """Callback to take action when teff slider changes"""
                teff = self.find_nearest_teff(new)
                if teff != old:
                    teff_message.text = "Closest grid point: {}".format(teff)
                    logg = logg_slider.value
                    metallicity = metallicity_slider.value
                    grid_point = (teff, logg, metallicity)
                    index = self.get_index(grid_point)

                    native_spec = self[index].normalize(percentile=95)
                    new_spec = native_spec.rotationally_broaden(
                        smoothing_slider.value
                    ).rv_shift(vz_slider.value)

                    spec_source.data = {
                        "native_wavelength": native_spec.wavelength.value,
                        "native_flux": native_spec.flux.value,
                        "wavelength": new_spec.wavelength.value,
                        "flux": new_spec.flux.value,
                    }

                else:
                    pass

            def update_upon_metallicity_selection(attr, old, new):
                """Callback to take action when teff slider changes"""
                metallicity = self.find_nearest_metallicity(new)
                if metallicity != old:
                    teff = self.find_nearest_teff(teff_slider.value)
                    logg = logg_slider.value
                    grid_point = (teff, logg, metallicity)
                    index = self.get_index(grid_point)

                    native_spec = self[index].normalize(percentile=95)
                    new_spec = native_spec.rotationally_broaden(
                        smoothing_slider.value
                    ).rv_shift(vz_slider.value)

                    spec_source.data = {
                        "native_wavelength": native_spec.wavelength.value,
                        "native_flux": native_spec.flux.value,
                        "wavelength": new_spec.wavelength.value,
                        "flux": new_spec.flux.value,
                    }

                else:
                    pass

            def update_upon_logg_selection(attr, old, new):
                """Callback to take action when logg slider changes"""
                teff = self.find_nearest_teff(teff_slider.value)
                metallicity = self.find_nearest_metallicity(metallicity_slider.value)
                grid_point = (teff, new, metallicity)
                index = self.get_index(grid_point)

                native_spec = self[index].normalize(percentile=95)
                new_spec = native_spec.rotationally_broaden(
                    smoothing_slider.value
                ).rv_shift(vz_slider.value)

                spec_source.data = {
                    "native_wavelength": native_spec.wavelength.value,
                    "native_flux": native_spec.flux.value,
                    "wavelength": new_spec.wavelength.value,
                    "flux": new_spec.flux.value,
                }

            def go_right_by_one():
                """Step forward in time by a single cadence"""
                current_index = np.abs(self.teff_points - teff_slider.value).argmin()
                new_index = current_index + 1
                if new_index <= (len(self.teff_points) - 1):
                    teff_slider.value = self.teff_points[new_index]

            def go_left_by_one():
                """Step back in time by a single cadence"""
                current_index = np.abs(self.teff_points - teff_slider.value).argmin()
                new_index = current_index - 1
                if new_index >= 0:
                    teff_slider.value = self.teff_points[new_index]

            r_button.on_click(go_right_by_one)
            l_button.on_click(go_left_by_one)
            smoothing_slider.on_change("value", update_upon_smooth)
            vz_slider.on_change("value", update_upon_vz)
            teff_slider.on_change("value", update_upon_teff_selection)
            logg_slider.on_change("value", update_upon_logg_selection)
            metallicity_slider.on_change("value", update_upon_metallicity_selection)

            sp1, sp2, sp3, sp4 = (
                Spacer(width=5),
                Spacer(width=10),
                Spacer(width=20),
                Spacer(width=100),
            )

            widgets_and_figures = layout(
                [fig],
                [l_button, sp1, r_button, sp2, teff_slider, sp3, teff_message],
                [sp4, logg_slider],
                [sp4, metallicity_slider],
                [sp4, smoothing_slider],
                [sp4, vz_slider],
            )
            doc.add_root(widgets_and_figures)

        output_notebook(verbose=False, hide_banner=True)
        return show(create_interact_ui, notebook_url=notebook_url)



import setuptools


setuptools.setup(
    name="gollum",
    version="0.2.0",
    author="gully",
    author_email="igully@gmail.com",
    description="A Python package for working with precomputed synthetic spectral models",
    long_description="A Python package for working with precomputed synthetic spectral models such as PHOENIX and Sonora-Bobcat",
    long_description_content_type="text/markdown",
    url="https://github.com/BrownDwarf/gollum",
    install_requires=[
        "numpy",
        "scipy",
        "astropy",
        "specutils>=1.5",
        "importlib_resources",
        "pandas",
        "matplotlib",
        "pyyaml",
    ],
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    package_data={
        # If any package contains *.csv files, include them:
        "": ["*.csv"]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",  # astropy 5 and up now requires Python 3.8+, sorry!
)



r"""
Precomputed Spectrum
-----------------

A container for a single precomputed synthetic model spectrum at a single grid-point, with wavelength and flux :math:`F(\lambda)`.

PrecomputedSpectrum
###############
"""

import warnings
import logging
import numpy as np
import astropy
from astropy import units as u
from astropy.modeling.physical_models import BlackBody
import specutils

import matplotlib.pyplot as plt
import copy

from scipy.ndimage import gaussian_filter1d
from scipy.interpolate import UnivariateSpline, interp1d
from scipy.signal import find_peaks, wavelets
from specutils.manipulation import LinearInterpolatedResampler
from specutils.fitting import fit_generic_continuum
from gollum.utilities import apply_numpy_mask


log = logging.getLogger(__name__)

#  See Issue: https://github.com/astropy/specutils/issues/779
warnings.filterwarnings(
    "ignore", category=astropy.utils.exceptions.AstropyDeprecationWarning
)
# See Issue: https://github.com/astropy/specutils/issues/800
warnings.filterwarnings("ignore", category=RuntimeWarning)

with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    from specutils import Spectrum1D
    from specutils import SpectrumList


class PrecomputedSpectrum(Spectrum1D):
    r"""
    An abstract container base class for a Precomputed spectrum

    """

    def __init__(self, *args, **kwargs):

        # Todo, we could put the wavelength limits in here.

        super().__init__(*args, **kwargs)

    def apply_boolean_mask(self, mask):
        """Apply a boolean mask to the spectrum

        Parameters
        ----------
        mask: boolean mask, typically a numpy array
            The boolean mask with numpy-style masking: True means "keep" that index and
            False means discard that index
        """

        spec = apply_numpy_mask(self, mask)

        return spec

    def normalize(self, percentile=None):
        """Normalize spectrum by its median value

        Args:
            percentile: The percentile to which the spectrum will be normalized
                (default: 50th percentile)

        Returns
        -------
        normalized_spec : (PrecomputedSpectrum)
            Normalized Spectrum
        """
        spec = self._copy(
            spectral_axis=self.wavelength.value * self.wavelength.unit, wcs=None
        )
        if percentile is None:
            scalar_flux = np.nanmedian(spec.flux.value) * spec.flux.unit
        else:
            scalar_flux = np.nanpercentile(spec.flux.value, percentile) * spec.flux.unit

        return spec.divide(scalar_flux, handle_meta="first_found")

    def rotationally_broaden(self, vsini, u1=0.0, u2=0.0):
        r"""Rotationally broaden the spectrum for a given :math:`v\sin{i}`
        Implementation inspired by https://github.com/HajimeKawahara/exojax

        Known limitation: If the wavelength sampling changes with wavelength,
          the convolution becomes inaccurate.  It may be better to FFT,
          following Starfish.

        Args:
            vsini: :math:`v\sin{i}` in units of km/s
            u1: Limb-darkening coefficient 1
            u2: Limb-darkening coefficient 2

        Returns
        -------
        broadened_spec : (PrecomputedSpectrum)
            Rotationally Broadened Spectrum
        """
        lam0 = np.median(self.wavelength.value)
        velocity_grid = 299792.458 * (self.wavelength.value - lam0) / lam0
        x = velocity_grid / vsini
        x2 = x * x
        kernel = np.where(
            x2 < 1.0,
            np.pi / 2.0 * u1 * (1.0 - x2)
            - 2.0 / 3.0 * np.sqrt(1.0 - x2) * (-3.0 + 3.0 * u1 + u2 * 2.0 * u2 * x2),
            0.0,
        )
        kernel = kernel / np.sum(kernel, axis=0)
        positive_elements = kernel > 0
        if positive_elements.any():
            kernel = kernel[positive_elements]
            convolved_flux = (
                np.convolve(self.flux.value, kernel, mode="same") * self.flux.unit
            )
            return self._copy(flux=convolved_flux)
        else:
            return self

    def instrumental_broaden(self, resolving_power=55_000):
        r"""Instrumentally broaden the spectrum for a given instrumental resolution, R

        Known limitation: If the wavelength sampling changes with wavelength,
          the convolution becomes inaccurate.  It may be better to FFT,
          following Starfish.

        Args:
            resolving_power: Instrumental resolving power :math:`R = \frac{\lambda}{\delta \lambda}`

        Returns
        -------
        broadened_spec : (PrecomputedSpectrum)
            Instrumentally Broadened Spectrum
        """
        # TODO: I think we want the Nadarya-Watson estimator here instead
        angstroms_per_pixel = np.median(np.diff(self.wavelength.value))
        lam0 = np.median(self.wavelength.value)
        delta_lam = lam0 / resolving_power

        scale_factor = 2.355
        sigma = delta_lam / scale_factor / angstroms_per_pixel

        convolved_flux = gaussian_filter1d(self.flux.value, sigma) * self.flux.unit
        return self._copy(flux=convolved_flux)

    def rv_shift(self, rv):
        """Shift the spectrum by a radial velocity, in units of km/s

        Args:
            rv: Radial velocity in units of km/s

        Returns
        -------
        shifted_spec : (PrecomputedSpectrum)
            RV Shifted Spectrum
        """
        try:
            output = copy.deepcopy(self)
            output.radial_velocity = rv * u.km / u.s
            return self._copy(
                spectral_axis=output.wavelength.value * output.wavelength.unit, wcs=None
            )
        except:
            log.error(
                "rv shift requires specutils version >= 1.2, you have: {}".format(
                    specutils.__version__
                )
            )
            raise

    def resample(self, target_spectrum):
        """Resample spectrum at the wavelength points of the other spectrum

        Args:
            target_spectrum: A Spectrum1D spectrum whose wavelength grid you seek to match

        Returns
        -------
        resampled_spec : (PrecomputedSpectrum)
            Resampled spectrum
        """
        fluxc_resample = LinearInterpolatedResampler()
        output = fluxc_resample(self, target_spectrum.wavelength)

        return self._copy(
            spectral_axis=output.wavelength.value * output.wavelength.unit,
            flux=output.flux,
            wcs=None,
        )

    def get_blackbody_spectrum(self, teff=None):
        """Get the blackbody spectrum associated with the input model"""
        if teff is None:
            # Try to look for a teff attribute
            if hasattr(self, "teff"):
                if self.teff is not None:
                    teff = self.teff
                else:
                    raise NotImplementedError(
                        "Your subclass may not have implemented the teff attribute yet."
                    )

        blackbody_model = BlackBody(temperature=teff * u.Kelvin)
        blackbody_flux_per_sr = blackbody_model(self.wavelength)

        blackbody_flux_per_Hz = blackbody_flux_per_sr * np.pi * u.steradian

        flux_unit = self.flux.unit
        normalize = False
        if flux_unit == u.dimensionless_unscaled:
            if "native_flux_unit" in self.meta.keys():
                flux_unit = self.meta["native_flux_unit"]
                normalize = True
            else:
                raise NotImplementedError(
                    "We do not yet support inferring units for this subclass"
                )

        blackbody_flux = blackbody_flux_per_Hz.to(
            flux_unit, equivalencies=u.spectral_density(self.wavelength)
        )

        output = PrecomputedSpectrum(flux=blackbody_flux, spectral_axis=self.wavelength)
        if normalize:
            return output.normalize()
        else:
            return output

    def divide_by_blackbody(self, teff=None):
        """Divide the spectrum by a blackbody

        Args:
            target_spectrum: Spectrum1D
                A Spectrum1D spectrum whose flux tilt you seek to match
            return_model: (bool)
                Whether or not to return the model

        Returns
        -------
        tilted_spec : (PrecomputedSpectrum)
            Tilted spectrum
        """
        return self.divide(self.get_blackbody_spectrum(teff=teff), handle_meta="ff")

    def tilt_to_data(self, target_spectrum, return_model=False):
        """Tilt the template towards a data spectrum by fitting and dividing by a low-order polynomial

        Args:
            target_spectrum: Spectrum1D
                A Spectrum1D spectrum whose flux tilt you seek to match
            return_model: (bool)
                Whether or not to return the model

        Returns
        -------
        tilted_spec : (PrecomputedSpectrum)
            Tilted spectrum
        """
        ## Assume the template is resampled exactly to the data...
        ratio = target_spectrum.divide(self, handle_meta="ff")
        g1_fit = fit_generic_continuum(ratio)
        y_continuum_fitted = g1_fit(target_spectrum.wavelength)

        tilted_spectrum = self.multiply(y_continuum_fitted)

        if return_model:
            return (tilted_spectrum, g1_fit)
        else:
            return tilted_spectrum

    def fit_continuum(
        self, pixel_distance=5001, polyorder=3, return_coeffs=False,
    ):
        """Finds the low frequency continuum trend using scipy's find_peaks filter
        and linear algebra.

        Parameters
        ----------
        pixel_distance : int
            The minimum separation between peaks, in pixels.  Default = 5001 pixels
        polyorder : int
            The order of the polynomial used to fit the peaks
        return_coeffs : bool
            If `True`, the method will return a tuple of two elements
            (trend_spec, trend_coeffs) where trend_spec is the fitted trend.
        Returns
        -------
        continuum_spec : `PrecomputedSpectrum`
            New Spectrum object representing a fit of the continuum.
        If ``return_coeffs`` is set to ``True``, this method will also return:
        trend_coeffs : np.array
            New vector of polynomial coefficients containing to reproduce the trend.
        """

        x_vector = self.wavelength.value
        y_vector = self.flux.value

        if pixel_distance > len(x_vector):
            raise IndexError(
                "Your pixel_distance is larger than the spectrum."
                " Please provide a smaller pixel distance."
            )
        peak_inds, _ = find_peaks(y_vector, distance=pixel_distance)

        x_peaks = x_vector[peak_inds]
        y_peaks = y_vector[peak_inds]

        A_matrix = np.vander(x_peaks, polyorder)
        A_full = np.vander(x_vector, polyorder)

        ATA = np.dot(A_matrix.T, A_matrix)

        coeffs = np.linalg.solve(ATA, np.dot(A_matrix.T, y_peaks))
        y_full = np.dot(coeffs, A_full.T)

        spec_out = self._copy(flux=y_full * self.flux.unit)

        if return_coeffs:
            return (spec_out, coeffs)
        else:
            return spec_out

    def to_pandas(self):
        """Export the spectrum to a pandas dataframe"""
        try:
            import pandas as pd
        except ImportError:
            log.error("The to_pandas method requires the optional dependency pandas")
        if self.mask is not None:
            mask = self.mask
        else:
            mask = np.zeros(len(self.wavelength.value), dtype=int)

        return pd.DataFrame(
            {
                "wavelength": self.wavelength.value,
                "flux": self.flux.value,
                "mask": mask,
            }
        )

    def plot(self, ax=None, ylo=0.6, yhi=1.2, figsize=(10, 4), **kwargs):
        """Plot a quick look of the spectrum"

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be generated.
        ylo : scalar
            Lower limit of the y axis
        yhi : scalar
            Upper limit of the y axis
        figsize : tuple
            The figure size for the plot
        label : str
            The legend label to for plt.legend()

        Returns
        -------
        ax : (`~matplotlib.axes.Axes`)
            The axis to display and/or modify
        """
        if ax is None:
            fig, ax = plt.subplots(1, figsize=figsize)
            ax.set_ylim(ylo, yhi)
            ax.set_xlabel(r"$\lambda \;(\AA)$")
            ax.set_ylabel("Flux")
            ax.step(self.wavelength, self.flux, where="mid", **kwargs)
        else:
            ax.step(self.wavelength, self.flux, where="mid", **kwargs)

        return ax



import copy
from specutils.spectra import Spectrum1D


def apply_numpy_mask(spec, mask):
    """Applies a boolean mask to an input spectrum, numpy-style (True=Keep, False=Discard)


    Parameters
    ----------
    spec: Spectrum1D-like object
        Object storing spectrum
    mask: boolean mask, typically a numpy array
        The boolean mask with numpy-style masking: True means "keep" that index and False means discard that index
    """

    assert isinstance(spec, Spectrum1D), "Input must be a specutils Spectrum1D object"

    assert mask.sum() > 0, "The masked spectrum must have at least one pixel remaining"

    if len(mask) != len(spec.spectral_axis.value):
        raise IndexError(
            "Your boolean mask has {} entries and your spectrum has {} pixels.  "
            " The boolean mask should have the same shape as the spectrum."
        )

    if spec.mask is not None:
        mask_out = spec.mask[mask]
    else:
        mask_out = None

    if spec.meta is not None:
        meta_out = copy.deepcopy(spec.meta)
    else:
        meta_out = None

    return spec.__class__(
        spectral_axis=spec.wavelength.value[mask] * spec.wavelength.unit,
        flux=spec.flux[mask],
        mask=mask_out,
        wcs=None,
        meta=meta_out,
    )
